"""
Some basic dicts and field-in-a-dict manipulation helpers.
"""
import collections.abc
from typing import Any, Union, Mapping, Tuple, List, Text, Iterable, Optional

FieldPath = Tuple[str, ...]
FieldSpec = Union[None, Text, FieldPath, List[str]]

_UNSET = object()


def parse_field(
        field: FieldSpec,
) -> FieldPath:
    """
    Convert any field into a tuple of nested sub-fields.
    """
    if field is None:
        return tuple()
    elif isinstance(field, str):
        return tuple(field.split('.'))
    elif isinstance(field, (list, tuple)):
        return tuple(field)
    else:
        raise ValueError(f"Field must be either a str, or a list/tuple. Got {field!r}")


def resolve(
        d: Mapping,
        field: FieldSpec,
        default: Any = _UNSET,
):
    """
    Retrieve a nested sub-field from a dict.
    """
    path = parse_field(field)
    try:
        result = d
        for key in path:
            result = result[key]
        return result
    except KeyError:
        if default is _UNSET:
            raise
        else:
            return default


def walk(
        objs,
        nested: Optional[Iterable[FieldSpec]] = None,
):
    """
    Iterate over one or many dicts (and sub-dicts recursively).
    """
    if objs is None:
        return
    elif isinstance(objs, collections.abc.Mapping):
        yield objs
        for subfield in (nested if nested is not None else []):
            try:
                yield resolve(objs, parse_field(subfield))
            except KeyError:
                pass
    elif isinstance(objs, collections.abc.Iterable):
        for obj in objs:
            yield from walk(obj, nested=nested)
    else:
        yield objs  # NB: not a mapping, no nested sub-fields.



import kubernetes
import pytest
from asynctest import call

from kopf.k8s.fetching import read_crd


def test_when_present(client_mock, resource):
    apicls_mock = client_mock.ApiextensionsV1beta1Api
    readfn_mock = apicls_mock.return_value.read_custom_resource_definition
    result_mock = readfn_mock.return_value

    crd = read_crd(resource=resource)
    assert crd is result_mock

    assert apicls_mock.called
    assert apicls_mock.call_count == 1
    assert readfn_mock.called
    assert readfn_mock.call_count == 1
    assert readfn_mock.call_args_list == [
        call(name=f'{resource.plural}.{resource.group}')
    ]


@pytest.mark.parametrize('status', [403, 404])
def test_when_absent_with_no_default(client_mock, resource, status):
    error = kubernetes.client.rest.ApiException(status=status)
    apicls_mock = client_mock.ApiextensionsV1beta1Api
    apicls_mock.return_value.read_custom_resource_definition.side_effect = error

    with pytest.raises(kubernetes.client.rest.ApiException) as e:
        read_crd(resource=resource)
    assert e.value.status == status


@pytest.mark.parametrize('default', [None, object()], ids=['none', 'object'])
@pytest.mark.parametrize('status', [403, 404])
def test_when_absent_with_default(client_mock, resource, default, status):
    error = kubernetes.client.rest.ApiException(status=status)
    apicls_mock = client_mock.ApiextensionsV1beta1Api
    apicls_mock.return_value.read_custom_resource_definition.side_effect = error

    crd = read_crd(resource=resource, default=default)
    assert crd is default


@pytest.mark.parametrize('status', [400, 401, 500, 666])
def test_raises_api_error_despite_default(client_mock, resource, status):
    error = kubernetes.client.rest.ApiException(status=status)
    apicls_mock = client_mock.ApiextensionsV1beta1Api
    apicls_mock.return_value.read_custom_resource_definition.side_effect = error

    with pytest.raises(kubernetes.client.rest.ApiException) as e:
        read_crd(resource=resource, default=object())
    assert e.value.status == status



import pytest
from asynctest import call, ANY

from kopf.k8s.events import post_event


def test_posting(client_mock):
    result = object()
    apicls_mock = client_mock.CoreV1Api
    apicls_mock.return_value.create_namespaced_event.return_value = result
    postfn_mock = apicls_mock.return_value.create_namespaced_event

    obj = {'apiVersion': 'group/version',
           'kind': 'kind',
           'metadata': {'namespace': 'ns',
                        'name': 'name',
                        'uid': 'uid'}}
    post_event(obj=obj, type='type', reason='reason', message='message')

    assert postfn_mock.called
    assert postfn_mock.call_count == 1
    assert postfn_mock.call_args_list == [call(
        namespace='ns',  # same as the object's namespace
        body=ANY,
    )]

    event = postfn_mock.call_args_list[0][1]['body']
    assert event.type == 'type'
    assert event.reason == 'reason'
    assert event.message == 'message'
    assert event.source.component == 'kopf'
    assert event.involved_object['apiVersion'] == 'group/version'
    assert event.involved_object['kind'] == 'kind'
    assert event.involved_object['namespace'] == 'ns'
    assert event.involved_object['name'] == 'name'
    assert event.involved_object['uid'] == 'uid'


def test_type_is_v1_not_v1beta1(client_mock):
    apicls_mock = client_mock.CoreV1Api
    postfn_mock = apicls_mock.return_value.create_namespaced_event

    obj = {'apiVersion': 'group/version',
           'kind': 'kind',
           'metadata': {'namespace': 'ns',
                        'name': 'name',
                        'uid': 'uid'}}
    post_event(obj=obj, type='type', reason='reason', message='message')

    event = postfn_mock.call_args_list[0][1]['body']
    assert isinstance(event, client_mock.V1Event)
    assert not isinstance(event, client_mock.V1beta1Event)


def test_api_errors_logged_but_suppressed(client_mock, assert_logs):
    error = client_mock.rest.ApiException('boo!')
    apicls_mock = client_mock.CoreV1Api
    apicls_mock.return_value.create_namespaced_event.side_effect = error
    postfn_mock = apicls_mock.return_value.create_namespaced_event

    obj = {'apiVersion': 'group/version',
           'kind': 'kind',
           'metadata': {'namespace': 'ns',
                        'name': 'name',
                        'uid': 'uid'}}
    post_event(obj=obj, type='type', reason='reason', message='message')

    assert postfn_mock.called
    assert_logs([
        "Failed to post an event.*boo!",
    ])


def test_regular_errors_escalate(client_mock):
    error = Exception('boo!')
    apicls_mock = client_mock.CoreV1Api
    apicls_mock.return_value.create_namespaced_event.side_effect = error

    obj = {'apiVersion': 'group/version',
           'kind': 'kind',
           'metadata': {'namespace': 'ns',
                        'name': 'name',
                        'uid': 'uid'}}

    with pytest.raises(Exception) as excinfo:
        post_event(obj=obj, type='type', reason='reason', message='message')

    assert excinfo.value is error


def test_message_is_cut_to_max_length(client_mock):
    result = object()
    apicls_mock = client_mock.CoreV1Api
    apicls_mock.return_value.create_namespaced_event.return_value = result
    postfn_mock = apicls_mock.return_value.create_namespaced_event

    obj = {'apiVersion': 'group/version',
           'kind': 'kind',
           'metadata': {'namespace': 'ns',
                        'name': 'name',
                        'uid': 'uid'}}
    message = 'start' + ('x' * 2048) + 'end'
    post_event(obj=obj, type='type', reason='reason', message=message)

    event = postfn_mock.call_args_list[0][1]['body']
    assert len(event.message) <= 1024  # max supported API message length
    assert '...' in event.message
    assert event.message.startswith('start')
    assert event.message.endswith('end')



"""
The routines to manipulate the handler progression over the event cycle.

Used to track which handlers are finished, which are not yet,
and how many retries were there.

There could be more than one low-level k8s watch-events per one actual
high-level kopf-event (a cause). The handlers are called at different times,
and the overall handling routine should persist the handler status somewhere.

The structure is this:

.. code-block: yaml

    metainfo: ...
    spec: ...
    status: ...
        kopf:
            progress:
                handler1:
                    started: 2018-12-31T23:59:59,999999
                    stopped: 2018-01-01T12:34:56,789000
                    success: true
                handler2:
                    started: 2018-12-31T23:59:59,999999
                    stopped: 2018-01-01T12:34:56,789000
                    failure: true
                    message: "Error message."
                handler3:
                    started: 2018-12-31T23:59:59,999999
                    retries: 30
                handler3/sub1:
                    started: 2018-12-31T23:59:59,999999
                    delayed: 2018-01-01T12:34:56,789000
                    retries: 10
                    message: "Not ready yet."
                handler3/sub2:
                    started: 2018-12-31T23:59:59,999999

* ``status.kopf.success`` are the handlers that succeeded (no re-execution).
* ``status.kopf.failure`` are the handlers that failed completely (no retries).
* ``status.kopf.delayed`` are the timestamps, until which these handlers sleep.
* ``status.kopf.retries`` are number of retries for succeeded, failed,
  and for the progressing handlers.

When the full event cycle is executed (possibly including multiple re-runs),
the whole ``status.kopf`` section is purged. The life-long persistence of status
is not intended: otherwise, multiple distinct causes will clutter the status
and collide with each other (especially critical for multiple updates).
"""

import collections.abc
import copy
import datetime


def is_started(*, body, handler):
    progress = body.get('status', {}).get('kopf', {}).get('progress', {})
    return handler.id in progress


def is_sleeping(*, body, handler):
    ts = get_awake_time(body=body, handler=handler)
    finished = is_finished(body=body, handler=handler)
    return not finished and ts is not None and ts > datetime.datetime.utcnow()


def is_awakened(*, body, handler):
    finished = is_finished(body=body, handler=handler)
    sleeping = is_sleeping(body=body, handler=handler)
    return bool(not finished and not sleeping)


def is_finished(*, body, handler):
    progress = body.get('status', {}).get('kopf', {}).get('progress', {})
    success = progress.get(handler.id, {}).get('success', None)
    failure = progress.get(handler.id, {}).get('failure', None)
    return bool(success or failure)


def get_start_time(*, body, patch, handler):
    progress = patch.get('status', {}).get('kopf', {}).get('progress', {})
    new_value = progress.get(handler.id, {}).get('started', None)
    progress = body.get('status', {}).get('kopf', {}).get('progress', {})
    old_value = progress.get(handler.id, {}).get('started', None)
    value = new_value or old_value
    return None if value is None else datetime.datetime.fromisoformat(value)


def get_awake_time(*, body, handler):
    progress = body.get('status', {}).get('kopf', {}).get('progress', {})
    value = progress.get(handler.id, {}).get('delayed', None)
    return None if value is None else datetime.datetime.fromisoformat(value)


def get_retry_count(*, body, handler):
    progress = body.get('status', {}).get('kopf', {}).get('progress', {})
    return progress.get(handler.id, {}).get('retries', None) or 0


def set_start_time(*, body, patch, handler):
    progress = patch.setdefault('status', {}).setdefault('kopf', {}).setdefault('progress', {})
    progress.setdefault(handler.id, {}).update({
        'started': datetime.datetime.utcnow().isoformat(),
    })


def set_awake_time(*, body, patch, handler, delay=None):
    if delay is not None:
        ts = datetime.datetime.utcnow() + datetime.timedelta(seconds=delay)
        ts = ts.isoformat()
    else:
        ts = None
    progress = patch.setdefault('status', {}).setdefault('kopf', {}).setdefault('progress', {})
    progress.setdefault(handler.id, {}).update({
        'delayed': ts,
    })


def set_retry_time(*, body, patch, handler, delay=None):
    retry = get_retry_count(body=body, handler=handler)
    progress = patch.setdefault('status', {}).setdefault('kopf', {}).setdefault('progress', {})
    progress.setdefault(handler.id, {}).update({
        'retries': retry + 1,
    })
    set_awake_time(body=body, patch=patch, handler=handler, delay=delay)


def store_failure(*, body, patch, handler, exc):
    retry = get_retry_count(body=body, handler=handler)
    progress = patch.setdefault('status', {}).setdefault('kopf', {}).setdefault('progress', {})
    progress.setdefault(handler.id, {}).update({
        'stopped': datetime.datetime.utcnow().isoformat(),
        'failure': True,
        'retries': retry + 1,
        'message': f'{exc}',
    })


def store_success(*, body, patch, handler, result=None):
    retry = get_retry_count(body=body, handler=handler)
    progress = patch.setdefault('status', {}).setdefault('kopf', {}).setdefault('progress', {})
    progress.setdefault(handler.id, {}).update({
        'stopped': datetime.datetime.utcnow().isoformat(),
        'success': True,
        'retries': retry + 1,
        'message': None,
    })
    store_result(patch=patch, handler=handler, result=result)


def store_result(*, patch, handler, result=None):
    if result is None:
        pass
    elif isinstance(result, collections.abc.Mapping):
        # TODO: merge recursively (patch-merge), do not overwrite the keys if they are present.
        patch.setdefault('status', {}).setdefault(handler.id, {}).update(result)
    else:
        # TODO? Fail if already present?
        patch.setdefault('status', {})[handler.id] = copy.deepcopy(result)


def purge_progress(*, body, patch):
    patch.setdefault('status', {}).setdefault('kopf', {})['progress'] = None



import asyncio
import re
import time

import asynctest
import pytest
import pytest_mock

from kopf.reactor.registries import Resource


# Make all tests in this directory and below asyncio-compatible by default.
def pytest_collection_modifyitems(items):
    for item in items:
        if asyncio.iscoroutinefunction(item.function):
            item.add_marker('asyncio')


# Substitute the regular mock with the async-aware mock in the `mocker` fixture.
@pytest.fixture(scope='session', autouse=True)
def enforce_asyncio_mocker():
    pytest_mock._get_mock_module = lambda config: asynctest


@pytest.fixture()
def resource():
    """ The resource used in the tests. Usually mocked, so it does not matter. """
    return Resource('zalando.org', 'v1', 'kopfexamples')

#
# Helpers for the timing checks.
#

@pytest.fixture()
def timer():
    return Timer()


class Timer(object):
    """
    A helper context manager to measure the time of the code-blocks.
    Also, supports direct comparison with time-deltas and the numbers of seconds.

    Usage:

        with Timer() as timer:
            do_something()
            print(f"Executing for {timer.seconds}s already.")
            do_something_else()

        print(f"Executed in {timer.seconds}s.")
        assert timer < 5.0
    """

    def __init__(self):
        super().__init__()
        self._ts = None
        self._te = None

    @property
    def seconds(self):
        if self._ts is None:
            return None
        elif self._te is None:
            return time.perf_counter() - self._ts
        else:
            return self._te - self._ts

    def __repr__(self):
        status = 'new' if self._ts is None else 'running' if self._te is None else 'finished'
        return f'<Timer: {self.seconds}s ({status})>'

    def __enter__(self):
        self._ts = time.perf_counter()
        self._te = None
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._te = time.perf_counter()

    def __int__(self):
        return int(self.seconds)

    def __float__(self):
        return float(self.seconds)

#
# Helpers for the logging checks.
#

@pytest.fixture()
def assert_logs(caplog):
    """
    A function to assert the logs are present (by pattern).

    The listed message patterns MUST be present, in the order specified.
    Some other log messages can also be present, but they are ignored.
    """
    def assert_logs_fn(patterns, prohibited=[], strict=False):
        __traceback_hide__ = True
        remaining_patterns = list(patterns)
        for message in caplog.messages:
            # The expected pattern is at position 0.
            # Looking-ahead: if one of the following patterns matches, while the
            # 0th does not, then the log message is missing, and we fail the test.
            for idx, pattern in enumerate(remaining_patterns):
                m = re.search(pattern, message)
                if m:
                    if idx == 0:
                        remaining_patterns[:1] = []
                        break  # out of `remaining_patterns` cycle
                    else:
                        skipped_patterns = remaining_patterns[:idx]
                        raise AssertionError(f"Few patterns were skipped: {skipped_patterns!r}")
                elif strict:
                    raise AssertionError(f"Unexpected log message: {message!r}")

            # Check that the prohibited patterns do not appear in any message.
            for pattern in prohibited:
                m = re.search(pattern, message)
                if m:
                    raise AssertionError(f"Prohibited log pattern found: {message!r} ~ {pattern!r}")

        # If all patterns have been matched in order, we are done.
        # if some are left, but the messages are over, then we fail.
        if remaining_patterns:
            raise AssertionError(f"Few patterns were missed: {remaining_patterns!r}")

    return assert_logs_fn



import pydoc

import kubernetes.client.rest
import pytest
from asynctest import call

from kopf.k8s.fetching import make_list_fn


def test_when_present_clustered(client_mock, resource):
    result = object()
    apicls_mock = client_mock.CustomObjectsApi
    apicls_mock.return_value.list_cluster_custom_object.return_value = result
    apicls_mock.return_value.list_namespaced_custom_object.return_value = result
    sidefn_mock = apicls_mock.return_value.list_namespaced_custom_object
    mainfn_mock = apicls_mock.return_value.list_cluster_custom_object

    fn = make_list_fn(resource=resource, namespace=None)
    assert callable(fn)

    assert not sidefn_mock.called
    assert not mainfn_mock.called

    res = fn(opt1='val1', opt2=123)
    assert res is result

    assert sidefn_mock.call_count == 0
    assert mainfn_mock.call_count == 1
    assert mainfn_mock.call_args_list == [call(
        group=resource.group,
        version=resource.version,
        plural=resource.plural,
        opt1='val1', opt2=123,
    )]


def test_when_present_namespaced(client_mock, resource):
    result = object()
    apicls_mock = client_mock.CustomObjectsApi
    apicls_mock.return_value.list_cluster_custom_object.return_value = result
    apicls_mock.return_value.list_namespaced_custom_object.return_value = result
    sidefn_mock = apicls_mock.return_value.list_cluster_custom_object
    mainfn_mock = apicls_mock.return_value.list_namespaced_custom_object

    fn = make_list_fn(resource=resource, namespace='ns1')
    assert callable(fn)

    assert not sidefn_mock.called
    assert not mainfn_mock.called

    res = fn(opt1='val1', opt2=123)
    assert res is result

    assert sidefn_mock.call_count == 0
    assert mainfn_mock.call_count == 1
    assert mainfn_mock.call_args_list == [call(
        group=resource.group,
        version=resource.version,
        plural=resource.plural,
        namespace='ns1',
        opt1='val1', opt2=123,
    )]


@pytest.mark.parametrize('namespace', [None, 'ns1'], ids=['without-namespace', 'with-namespace'])
@pytest.mark.parametrize('status', [400, 401, 403, 404, 500, 666])
def test_raises_api_error(client_mock, resource, namespace, status):
    error = kubernetes.client.rest.ApiException(status=status)
    apicls_mock = client_mock.CustomObjectsApi
    apicls_mock.return_value.list_cluster_custom_object.side_effect = error
    apicls_mock.return_value.list_namespaced_custom_object.side_effect = error

    fn = make_list_fn(resource=resource, namespace=namespace)
    with pytest.raises(kubernetes.client.rest.ApiException) as e:
        fn(opt1='val1', opt2=123)
    assert e.value.status == status


@pytest.mark.parametrize('namespace', [None, 'ns1'], ids=['without-namespace', 'with-namespace'])
def test_docstrings_are_preserved(client_mock, resource, namespace):
    # Docstrings are important! Kubernetes client uses them to guess
    # the returned object types and the parameters type.
    docstring = """some doc \n :return: sometype"""

    apicls_mock = client_mock.CustomObjectsApi
    apicls_mock.return_value.list_cluster_custom_object.__doc__ = docstring
    apicls_mock.return_value.list_namespaced_custom_object.__doc__ = docstring

    fn = make_list_fn(resource=resource, namespace=namespace)
    fn_docstring = pydoc.getdoc(fn)  # same as k8s client does this
    assert isinstance(fn_docstring, str)
    assert ':return: sometype' in docstring  # it will be reformatted



"""
A registry of the handlers, attached to the resources or events.

The global registry is populated by the `kopf.on` decorators, and is used
to register the resources being watched and handled, and to attach
the handlers to the specific causes (create/update/delete/field-change).

The simple registry is part of the global registry (for each individual
resource), and also used for the sub-handlers within a top-level handler.

Both are used in the `kopf.reactor.handling` to retrieve the list
of the handlers to be executed on each reaction cycle.
"""
import abc
import collections
import functools
from types import FunctionType, MethodType
from typing import MutableMapping


# An immutable reference to a custom resource definition.
Resource = collections.namedtuple('Resource', 'group version plural')

# A registered handler (function + event meta info).
Handler = collections.namedtuple('Handler', 'fn id event field timeout initial')


class BaseRegistry(metaclass=abc.ABCMeta):
    """
    A registry stores the handlers and provides them to the reactor.
    """

    def get_cause_handlers(self, cause):
        return list(self._deduplicated(self.iter_cause_handlers(cause=cause)))

    @abc.abstractmethod
    def iter_cause_handlers(self, cause):
        pass

    def get_event_handlers(self, resource, event):
        return list(self._deduplicated(self.iter_event_handlers(resource=resource, event=event)))

    @abc.abstractmethod
    def iter_event_handlers(self, resource, event):
        pass

    @staticmethod
    def _deduplicated(handlers):
        """
        Yield the handlers deduplicated.

        The same handler function should not be invoked more than once for one
        single event/cause, even if it is registered with multiple decorators
        (e.g. different filtering criteria or different but same-effect causes).

        One of the ways how this could happen::

            @kopf.on.create(...)
            @kopf.on.resume(...)
            def fn(**kwargs): pass

        In normal cases, the function will be called either on resource creation
        or on operator restart for the pre-existing (already handled) resources.
        When a resource is created during the operator downtime, it is
        both creation and resuming at the same time: the object is new (not yet
        handled) **AND** it is detected as per-existing before operator start.
        But `fn()` should be called only once for this cause.
        """
        seen_ids = set()
        for handler in handlers:
            if id(handler.fn) in seen_ids:
                pass
            else:
                seen_ids.add(id(handler.fn))
                yield handler


class SimpleRegistry(BaseRegistry):
    """
    A simple registry is just a list of handlers, no grouping.
    """

    def __init__(self, prefix=None):
        super().__init__()
        self.prefix = prefix
        self._handlers = []  # [Handler, ...]

    def __bool__(self):
        return bool(self._handlers)

    def append(self, handler):
        self._handlers.append(handler)

    def register(self, fn, id=None, event=None, field=None, timeout=None, initial=None):

        if field is None:
            field = None  # for the non-field events
        elif isinstance(field, str):
            field = tuple(field.split('.'))
        elif isinstance(field, (list, tuple)):
            field = tuple(field)
        else:
            raise ValueError(f"Field must be either a str, or a list/tuple. Got {field!r}")

        id = id if id is not None else get_callable_id(fn)
        id = id if field is None else f'{id}/{".".join(field)}'
        id = id if self.prefix is None else f'{self.prefix}/{id}'
        handler = Handler(id=id, fn=fn, event=event, field=field, timeout=timeout, initial=initial)

        self.append(handler)
        return fn  # to be usable as a decorator too.

    def iter_cause_handlers(self, cause):
        fields = {field for _, field, _, _ in cause.diff or []}
        for handler in self._handlers:
            if handler.event is None or handler.event == cause.event:
                if handler.initial and not cause.initial:
                    pass  # ignore initial handlers in non-initial causes.
                elif handler.field:
                    if any(field[:len(handler.field)] == handler.field for field in fields):
                        yield handler
                else:
                    yield handler

    def iter_event_handlers(self, resource, event):
        for handler in self._handlers:
            yield handler


def get_callable_id(c):
    """ Get an reasonably good id of any commonly used callable. """
    if c is None:
        return None
    elif isinstance(c, functools.partial):
        return get_callable_id(c.func)
    elif hasattr(c, '__wrapped__'):  # @functools.wraps()
        return get_callable_id(c.__wrapped__)
    elif isinstance(c, FunctionType) and c.__name__ == '<lambda>':
        # The best we can do to keep the id stable across the process restarts,
        # assuming at least no code changes. The code changes are not detectable.
        line = c.__code__.co_firstlineno
        path = c.__code__.co_filename
        return f'lambda:{path}:{line}'
    elif isinstance(c, (FunctionType, MethodType)):
        return getattr(c, '__qualname__', getattr(c, '__name__', repr(c)))
    else:
        raise ValueError(f"Cannot get id of {c!r}.")


class GlobalRegistry(BaseRegistry):
    """
    A global registry is used for handling of the multiple resources.
    It is usually populated by the `@kopf.on...` decorators.
    """

    def __init__(self):
        super().__init__()
        self._cause_handlers: MutableMapping[Resource, SimpleRegistry] = {}
        self._event_handlers: MutableMapping[Resource, SimpleRegistry] = {}

    def register_cause_handler(self, group, version, plural, fn,
                               id=None, event=None, field=None, timeout=None, initial=None):
        """
        Register an additional handler function for the specific resource and specific event.
        """
        resource = Resource(group, version, plural)
        registry = self._cause_handlers.setdefault(resource, SimpleRegistry())
        registry.register(event=event, field=field, fn=fn, id=id, timeout=timeout, initial=initial)
        return fn  # to be usable as a decorator too.

    def register_event_handler(self, group, version, plural, fn, id=None):
        """
        Register an additional handler function for low-level events.
        """
        resource = Resource(group, version, plural)
        registry = self._event_handlers.setdefault(resource, SimpleRegistry())
        registry.register(fn=fn, id=id)
        return fn  # to be usable as a decorator too.

    @property
    def resources(self):
        """ All known resources in the registry. """
        return frozenset(self._cause_handlers) | frozenset(self._event_handlers)

    def has_cause_handlers(self, resource):
        resource_registry = self._cause_handlers.get(resource, None)
        return bool(resource_registry)

    def has_event_handlers(self, resource):
        resource_registry = self._event_handlers.get(resource, None)
        return bool(resource_registry)

    def iter_cause_handlers(self, cause):
        """
        Iterate all handlers that match this cause/event, in the order they were registered (even if mixed).
        """
        resource_registry = self._cause_handlers.get(cause.resource, None)
        if resource_registry is not None:
            yield from resource_registry.iter_cause_handlers(cause=cause)

    def iter_event_handlers(self, resource, event):
        """
        Iterate all handlers for the low-level events.
        """
        resource_registry = self._event_handlers.get(resource, None)
        if resource_registry is not None:
            yield from resource_registry.iter_event_handlers(resource=resource, event=event)


_default_registry = GlobalRegistry()


def get_default_registry() -> GlobalRegistry:
    """
    Get the default registry to be used by the decorators and the reactor
    unless the explicit registry is provided to them.
    """
    return _default_registry


def set_default_registry(registry: GlobalRegistry):
    """
    Set the default registry to be used by the decorators and the reactor
    unless the explicit registry is provided to them.
    """
    global _default_registry
    _default_registry = registry



"""
All the functions to write the Kubernetes events on the Kubernetes objects.

They are used internally in the handling routine to show the progress,
and can be used directly from the handlers to add arbitrary custom events.

The events look like this:

    kubectl describe -f myres.yaml
    ...
    TODO

"""
import sys

from kopf.k8s import events


# TODO: rename it it kopf.log()? kopf.events.log()? kopf.events.warn()?
def event(obj, *, type, reason, message=''):
    """
    Issue an event for the object.
    """
    if isinstance(obj, (list, tuple)):
        for item in obj:
            events.post_event(obj=item, type=type, reason=reason, message=message)
    else:
        events.post_event(obj=obj, type=type, reason=reason, message=message)


# Shortcuts for the only two officially documented event types as of now.
# However, any arbitrary strings can be used as an event type to the base function.
def info(obj, *, reason, message=''):
    return event(obj, reason=reason, message=message, type='Normal')


def warn(obj, *, reason, message=''):
    return event(obj, reason=reason, message=message, type='Warning')


def exception(obj, *, reason='', message='', exc=None):
    if exc is None:
        _, exc, _ = sys.exc_info()
    reason = reason if reason else type(exc).__name__
    message = f'{message} {exc}' if message else f'{exc}'
    return event(obj, reason=reason, message=message, type='Error')



import pytest
from kubernetes.client import V1Event as V1Event
from kubernetes.client import V1EventSource as V1EventSource
from kubernetes.client import V1ObjectMeta as V1ObjectMeta
from kubernetes.client import V1beta1Event as V1beta1Event
from kubernetes.client.rest import ApiException  # to avoid mocking it


# We do not test the Kubernetes client, so everything there should be mocked.
# Also, no external calls must be made under any circumstances. Hence, auto-use.
@pytest.fixture(autouse=True)
def client_mock(mocker):
    client_mock = mocker.patch('kubernetes.client')
    client_mock.rest.ApiException = ApiException  # to be raises and caught
    client_mock.V1Event = V1Event
    client_mock.V1beta1Event = V1beta1Event
    client_mock.V1EventSource = V1EventSource
    client_mock.V1ObjectMeta = V1ObjectMeta
    return client_mock



import kubernetes.client.rest
import pytest
from asynctest import call

from kopf.k8s.fetching import read_obj


def test_when_present_clustered(client_mock, resource):
    result = object()
    apicls_mock = client_mock.CustomObjectsApi
    apicls_mock.return_value.get_cluster_custom_object.return_value = result
    apicls_mock.return_value.get_namespaced_custom_object.return_value = result
    sidefn_mock = apicls_mock.return_value.get_namespaced_custom_object
    mainfn_mock = apicls_mock.return_value.get_cluster_custom_object

    crd = read_obj(resource=resource, namespace=None, name='name1')
    assert crd is result

    assert not sidefn_mock.called
    assert mainfn_mock.call_count == 1
    assert mainfn_mock.call_args_list == [call(
        group=resource.group,
        version=resource.version,
        plural=resource.plural,
        name='name1',
    )]


def test_when_present_namespaced(client_mock, resource):
    result = object()
    apicls_mock = client_mock.CustomObjectsApi
    apicls_mock.return_value.get_cluster_custom_object.return_value = result
    apicls_mock.return_value.get_namespaced_custom_object.return_value = result
    sidefn_mock = apicls_mock.return_value.get_cluster_custom_object
    mainfn_mock = apicls_mock.return_value.get_namespaced_custom_object

    crd = read_obj(resource=resource, namespace='ns1', name='name1')
    assert crd is result

    assert not sidefn_mock.called
    assert mainfn_mock.call_count == 1
    assert mainfn_mock.call_args_list == [call(
        group=resource.group,
        version=resource.version,
        plural=resource.plural,
        namespace='ns1',
        name='name1',
    )]


@pytest.mark.parametrize('namespace', [None, 'ns1'], ids=['without-namespace', 'with-namespace'])
@pytest.mark.parametrize('status', [404])
def test_when_absent_with_no_default(client_mock, resource, namespace, status):
    error = kubernetes.client.rest.ApiException(status=status)
    apicls_mock = client_mock.CustomObjectsApi
    apicls_mock.return_value.get_cluster_custom_object.side_effect = error
    apicls_mock.return_value.get_namespaced_custom_object.side_effect = error

    with pytest.raises(kubernetes.client.rest.ApiException) as e:
        read_obj(resource=resource, namespace=namespace, name='name1')
    assert e.value.status == status


@pytest.mark.parametrize('default', [None, object()], ids=['none', 'object'])
@pytest.mark.parametrize('namespace', [None, 'ns1'], ids=['without-namespace', 'with-namespace'])
@pytest.mark.parametrize('status', [404])
def test_when_absent_with_default(client_mock, resource, namespace, default, status):
    error = kubernetes.client.rest.ApiException(status=status)
    apicls_mock = client_mock.CustomObjectsApi
    apicls_mock.return_value.get_cluster_custom_object.side_effect = error
    apicls_mock.return_value.get_namespaced_custom_object.side_effect = error

    crd = read_obj(resource=resource, namespace=namespace, name='name1', default=default)
    assert crd is default


@pytest.mark.parametrize('namespace', [None, 'ns1'], ids=['without-namespace', 'with-namespace'])
@pytest.mark.parametrize('status', [400, 401, 403, 500, 666])
def test_raises_api_error_despite_default(client_mock, resource, namespace, status):
    error = kubernetes.client.rest.ApiException(status=status)
    apicls_mock = client_mock.CustomObjectsApi
    apicls_mock.return_value.get_cluster_custom_object.side_effect = error
    apicls_mock.return_value.get_namespaced_custom_object.side_effect = error

    with pytest.raises(kubernetes.client.rest.ApiException) as e:
        read_obj(resource=resource, namespace=namespace, name='name1', default=object())
    assert e.value.status == status



import datetime
import logging

import kubernetes.client.rest

logger = logging.getLogger(__name__)

MAX_MESSAGE_LENGTH = 1024
CUT_MESSAGE_INFIX = '...'


def post_event(*, obj, type, reason, message=''):
    """
    Issue an event for the object.
    """

    now = datetime.datetime.utcnow()
    namespace = obj['metadata']['namespace']

    # Prevent a common case of event posting errors but shortening the message.
    if len(message) > MAX_MESSAGE_LENGTH:
        infix = CUT_MESSAGE_INFIX
        prefix = message[:MAX_MESSAGE_LENGTH // 2 - (len(infix) // 2)]
        suffix = message[-MAX_MESSAGE_LENGTH // 2 + (len(infix) - len(infix) // 2):]
        message = f'{prefix}{infix}{suffix}'

    # Object reference - similar to the owner reference, but different.
    ref = dict(
        apiVersion=obj['apiVersion'],
        kind=obj['kind'],
        name=obj['metadata']['name'],
        uid=obj['metadata']['uid'],
        namespace=obj['metadata']['namespace'],
    )

    meta = kubernetes.client.V1ObjectMeta(
        namespace=namespace,
        generate_name='kopf-event-',
    )
    body = kubernetes.client.V1Event(
        metadata=meta,

        action='Action?',
        type=type,
        reason=reason,
        message=message,

        reporting_component='kopf',
        reporting_instance='dev',
        source=kubernetes.client.V1EventSource(component='kopf'),  # used in the "From" column in `kubectl describe`.

        involved_object=ref,

        first_timestamp=now.isoformat() + 'Z',  # '2019-01-28T18:25:03.000000Z' -- seen in `kubectl describe ...`
        last_timestamp=now.isoformat() + 'Z',  # '2019-01-28T18:25:03.000000Z' - seen in `kubectl get events`
        event_time=now.isoformat() + 'Z',  # '2019-01-28T18:25:03.000000Z'
    )

    try:
        api = kubernetes.client.CoreV1Api()
        api.create_namespaced_event(
            namespace=namespace,
            body=body,
        )
    except kubernetes.client.rest.ApiException as e:
        # Events are helpful but auxiliary, they should not fail the handling cycle.
        # Yet we want to notice that something went wrong (in logs).
        logger.warning("Failed to post an event. Ignoring and continuing. "
                       f"Error: {e!r}. "
                       f"Event: type={type!r}, reason={reason!r}, message={message!r}.")



import copy
import json

import pytest

from kopf.reactor.causation import CREATE, UPDATE, DELETE, NEW, NOOP, FREE, GONE
from kopf.reactor.causation import detect_cause
from kopf.structs.finalizers import FINALIZER
from kopf.structs.lastseen import LAST_SEEN_ANNOTATION

# Encoded at runtime, so that we do not make any assumptions on json formatting.
SPEC_DATA = {'spec': {'field': 'value'}}
SPEC_JSON = json.dumps((SPEC_DATA))
ALT_DATA = {'spec': {'field': 'other'}}
ALT_JSON = json.dumps((ALT_DATA))

#
# The following factors contribute to the detection of the cause
# (and we combine all of them with the matching & mismatching fixtures):
# * Finalizers (presence or absence).
# * Deletion timestamp (presence or absence).
# * Annotation with the last-seen state (presence or absence).
# * Annotation with the last-seen state (difference with the real state).
#

deleted_events = pytest.mark.parametrize('event', [
    pytest.param('DELETED'),
])

regular_events = pytest.mark.parametrize('event', [
    pytest.param('ADDED'),
    pytest.param('MODIFIED'),
    pytest.param('FORWARD-COMPATIBILITY-PSEUDO-EVENT', id='COMPAT'),
])


all_finalizers = pytest.mark.parametrize('finalizers', [
    pytest.param({}, id='no-finalizers'),
    pytest.param({'finalizers': [FINALIZER]}, id='own-finalizer'),
    pytest.param({'finalizers': ['irrelevant', 'another']}, id='other-finalizers'),
    pytest.param({'finalizers': ['irrelevant', FINALIZER, 'another']}, id='mixed-finalizers'),
])

our_finalizers = pytest.mark.parametrize('finalizers', [
    pytest.param({'finalizers': [FINALIZER]}, id='own-finalizer'),
    pytest.param({'finalizers': ['irrelevant', FINALIZER, 'another']}, id='mixed-finalizers'),
])

no_finalizers = pytest.mark.parametrize('finalizers', [
    pytest.param({}, id='no-finalizers'),
    pytest.param({'finalizers': ['irrelevant', 'another']}, id='other-finalizers'),
])


all_deletions = pytest.mark.parametrize('deletion_ts', [
    pytest.param({}, id='no-deletion-ts'),
    pytest.param({'deletionTimestamp': None}, id='empty-deletion-ts'),
    pytest.param({'deletionTimestamp': 'some'}, id='real-deletion-ts'),
])

real_deletions = pytest.mark.parametrize('deletion_ts', [
    pytest.param({'deletionTimestamp': 'some'}, id='real-deletion-ts'),
])

no_deletions = pytest.mark.parametrize('deletion_ts', [
    pytest.param({}, id='no-deletion-ts'),
    pytest.param({'deletionTimestamp': None}, id='empty-deletion-ts'),
])


all_lastseen = pytest.mark.parametrize('annotations', [
    pytest.param({}, id='no-annotations'),
    pytest.param({'annotations': {}}, id='no-last-seen'),
    pytest.param({'annotations': {LAST_SEEN_ANNOTATION: SPEC_JSON}}, id='matching-last-seen'),
    pytest.param({'annotations': {LAST_SEEN_ANNOTATION: SPEC_JSON}}, id='mismatching-last-seen'),
])

absent_lastseen = pytest.mark.parametrize('annotations', [
    pytest.param({}, id='no-annotations'),
    pytest.param({'annotations': {}}, id='no-last-seen'),
])

matching_lastseen = pytest.mark.parametrize('annotations', [
    pytest.param({'annotations': {LAST_SEEN_ANNOTATION: SPEC_JSON}}, id='matching-last-seen'),
])

mismatching_lastseen = pytest.mark.parametrize('annotations', [
    pytest.param({'annotations': {LAST_SEEN_ANNOTATION: ALT_JSON}}, id='mismatching-last-seen'),
])


@pytest.fixture
def content():
    return copy.deepcopy(SPEC_DATA)


#
# kwargs helpers -- to test them for all causes.
#

@pytest.fixture()
def kwargs():
    return dict(
        resource=object(),
        logger=object(),
        patch=object(),
    )

def check_kwargs(cause, kwargs):
    __traceback_hide__ = True
    assert cause.resource is kwargs['resource']
    assert cause.logger is kwargs['logger']
    assert cause.patch is kwargs['patch']


#
# The tests.
#

@all_finalizers
@all_deletions
@deleted_events
def test_for_gone(kwargs, event, finalizers, deletion_ts):
    event = {'type': event, 'object': {'metadata': {}}}
    event['object']['metadata'].update(finalizers)
    event['object']['metadata'].update(deletion_ts)
    cause = detect_cause(event=event, **kwargs)
    assert cause.event == GONE
    check_kwargs(cause, kwargs)


@no_finalizers
@real_deletions
@regular_events
def test_for_free(kwargs, event, finalizers, deletion_ts):
    event = {'type': event, 'object': {'metadata': {}}}
    event['object']['metadata'].update(finalizers)
    event['object']['metadata'].update(deletion_ts)
    cause = detect_cause(event=event, **kwargs)
    assert cause.event == FREE
    check_kwargs(cause, kwargs)


@our_finalizers
@real_deletions
@regular_events
def test_for_delete(kwargs, event, finalizers, deletion_ts):
    event = {'type': event, 'object': {'metadata': {}}}
    event['object']['metadata'].update(finalizers)
    event['object']['metadata'].update(deletion_ts)
    cause = detect_cause(event=event, **kwargs)
    assert cause.event == DELETE
    check_kwargs(cause, kwargs)


@no_finalizers
@no_deletions
@regular_events
def test_for_new(kwargs, event, finalizers, deletion_ts):
    event = {'type': event, 'object': {'metadata': {}}}
    event['object']['metadata'].update(finalizers)
    event['object']['metadata'].update(deletion_ts)
    cause = detect_cause(event=event, **kwargs)
    assert cause.event == NEW
    check_kwargs(cause, kwargs)


@absent_lastseen
@our_finalizers
@no_deletions
@regular_events
def test_for_create(kwargs, event, finalizers, deletion_ts, annotations, content):
    event = {'type': event, 'object': {'metadata': {}}}
    event['object'].update(content)
    event['object']['metadata'].update(finalizers)
    event['object']['metadata'].update(deletion_ts)
    event['object']['metadata'].update(annotations)
    cause = detect_cause(event=event, **kwargs)
    assert cause.event == CREATE
    check_kwargs(cause, kwargs)


@matching_lastseen
@our_finalizers
@no_deletions
@regular_events
def test_for_no_op(kwargs, event, finalizers, deletion_ts, annotations, content):
    event = {'type': event, 'object': {'metadata': {}}}
    event['object'].update(content)
    event['object']['metadata'].update(finalizers)
    event['object']['metadata'].update(deletion_ts)
    event['object']['metadata'].update(annotations)
    cause = detect_cause(event=event, **kwargs)
    assert cause.event == NOOP
    check_kwargs(cause, kwargs)


@mismatching_lastseen
@our_finalizers
@no_deletions
@regular_events
def test_for_update(kwargs, event, finalizers, deletion_ts, annotations, content):
    event = {'type': event, 'object': {'metadata': {}}}
    event['object'].update(content)
    event['object']['metadata'].update(finalizers)
    event['object']['metadata'].update(deletion_ts)
    event['object']['metadata'].update(annotations)
    cause = detect_cause(event=event, **kwargs)
    assert cause.event == UPDATE
    check_kwargs(cause, kwargs)

    # Diffs are tested elsewhere, but quickly check the absence of meta-fields.
    assert cause.diff
    assert len(cause.diff) == 1
    assert cause.diff[0][0] == 'change'
    assert cause.diff[0][1] == ('spec', 'field')
    assert cause.diff[0][2] == 'other'
    assert cause.diff[0][3] == 'value'



import pytest

from kopf.reactor.registries import Handler


def test_no_args():
    with pytest.raises(TypeError):
        Handler()


def test_all_args(mocker):
    fn = mocker.Mock()
    id = mocker.Mock()
    event = mocker.Mock()
    field = mocker.Mock()
    timeout = mocker.Mock()
    initial = mocker.Mock()
    handler = Handler(
        fn=fn,
        id=id,
        event=event,
        field=field,
        timeout=timeout,
        initial=initial,
    )
    assert handler.fn is fn
    assert handler.id is id
    assert handler.event is event
    assert handler.field is field
    assert handler.timeout is timeout
    assert handler.initial is initial



import functools

import kubernetes

_UNSET_ = object()


def read_crd(*, resource, default=_UNSET_):
    try:
        name = f'{resource.plural}.{resource.group}'
        api = kubernetes.client.ApiextensionsV1beta1Api()
        rsp = api.read_custom_resource_definition(name=name)
        return rsp
    except kubernetes.client.rest.ApiException as e:
        if e.status in [404, 403] and default is not _UNSET_:
            return default
        raise


def read_obj(*, resource, namespace=None, name=None, default=_UNSET_):
    try:
        if namespace is None:
            api = kubernetes.client.CustomObjectsApi()
            rsp = api.get_cluster_custom_object(
                group=resource.group,
                version=resource.version,
                plural=resource.plural,
                name=name,
            )
        else:
            api = kubernetes.client.CustomObjectsApi()
            rsp = api.get_namespaced_custom_object(
                group=resource.group,
                version=resource.version,
                plural=resource.plural,
                namespace=namespace,
                name=name,
            )
        return rsp
    except kubernetes.client.rest.ApiException as e:
        if e.status == 404 and default is not _UNSET_:
            return default
        raise


def list_objs(*, resource, namespace=None):
    """
    List the objects of specific resource type.

    The cluster-scoped call is used in two cases:

    * The resource itself is cluster-scoped, and namespacing makes not sense.
    * The operator serves all namespaces for the namespaced custom resource.

    Otherwise, the namespace-scoped call is used:

    * The resource is namespace-scoped AND operator is namespaced-restricted.
    """
    api = kubernetes.client.CustomObjectsApi()
    if namespace is None:
        rsp = api.list_cluster_custom_object(
            group=resource.group,
            version=resource.version,
            plural=resource.plural,
        )
    else:
        rsp = api.list_namespaced_custom_object(
            group=resource.group,
            version=resource.version,
            plural=resource.plural,
            namespace=namespace,
        )
    return rsp


def make_list_fn(*, resource, namespace=None):
    """
    Return a function to be called to receive the list of objects.
    Needed in that form for the API streaming calls (see watching.py).

    However, the returned function is already bound to the specified
    resource type, and requires no resource-identifying parameters.

    Docstrings are important! Kubernetes client uses them to guess
    the returned object types and the parameters type.
    Function wrapping does that: preserves the docstrings.
    """
    api = kubernetes.client.CustomObjectsApi()
    if namespace is None:
        @functools.wraps(api.list_cluster_custom_object)
        def list_fn(**kwargs):
            return api.list_cluster_custom_object(
                group=resource.group,
                version=resource.version,
                plural=resource.plural,
                **kwargs)
    else:
        @functools.wraps(api.list_cluster_custom_object)
        def list_fn(**kwargs):
            return api.list_namespaced_custom_object(
                group=resource.group,
                version=resource.version,
                plural=resource.plural,
                namespace=namespace,
                **kwargs)
    return list_fn



import pytest

from kopf.testing import KopfRunner


@pytest.fixture(autouse=True)
def no_config_needed(mocker):
    mocker.patch('kubernetes.config.load_incluster_config')
    mocker.patch('kubernetes.config.load_kube_config')
    mocker.patch('kubernetes.client.CoreApi')  # for login self-check


def test_command_invocation_works():
    with KopfRunner(['--help']) as runner:
        pass
    assert runner.exc_info
    assert runner.exc_info[0] is SystemExit
    assert runner.exc_info[1].code == 0
    assert runner.exit_code == 0
    assert runner.exception is None
    assert runner.output.startswith("Usage:")
    assert runner.stdout.startswith("Usage:")
    assert runner.stdout_bytes.startswith(b"Usage:")
    # Note: stderr is not captured, it is mixed with stdout.


def test_exception_from_runner_suppressed_with_no_reraise():
    with KopfRunner(['run', 'non-existent.py', '--standalone'], reraise=False) as runner:
        pass
    assert runner.exception is not None
    assert isinstance(runner.exception, FileNotFoundError)


def test_exception_from_runner_escalates_with_reraise():
    with pytest.raises(FileNotFoundError):
        with KopfRunner(['run', 'non-existent.py', '--standalone'], reraise=True):
            pass


def test_exception_from_runner_escalates_by_default():
    with pytest.raises(FileNotFoundError):
        with KopfRunner(['run', 'non-existent.py', '--standalone']):
            pass


@pytest.mark.parametrize('kwargs',[
    dict(reraise=False),
    dict(reraise=True),
    dict(),
], ids=['reraise=False', 'reraise=True', 'no-reraise'])
def test_exception_from_invoke_escalates(mocker, kwargs):
    class SampleError(Exception): pass
    mocker.patch('click.testing.CliRunner.invoke', side_effect=SampleError)

    with pytest.raises(SampleError):
        with KopfRunner(['run', 'non-existent.py', '--standalone'], **kwargs):
            pass


def test_wrong_command_fails():
    with pytest.raises(SystemExit) as e:
        with KopfRunner(['unexistent-command']):
            pass
    assert e.value.code == 2


def test_absent_file_fails():
    with pytest.raises(FileNotFoundError):
        with KopfRunner(['run', 'non-existent.py', '--standalone']):
            pass


def test_bad_syntax_file_fails(tmpdir):
    path = tmpdir.join('handlers.py')
    path.write("""This is a Python syntax error!""")
    with pytest.raises((IndentationError, SyntaxError)):
        with KopfRunner(['run', str(path), '--standalone']):
            pass



import glob
import os.path
import pathlib
import subprocess

import pytest

import kopf

root_dir = os.path.relpath(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
examples = sorted(glob.glob(os.path.join(root_dir, 'examples/*/')))
assert examples  # if empty, it is just the detection failed


@pytest.fixture(params=examples)
def exampledir(request):
    return pathlib.Path(request.param)


@pytest.fixture(scope='session', autouse=True)
def autologin():
    kopf.config.login()  # or anything like that; it is not a unit-under-test


@pytest.fixture()
def with_crd():
    subprocess.run("kubectl apply -f examples/crd.yaml", shell=True, check=True)


@pytest.fixture()
def with_peering():
    subprocess.run("kubectl apply -f peering.yaml", shell=True, check=True)


@pytest.fixture()
def no_crd():
    subprocess.run("kubectl delete customresourcedefinition kopfexamples.zalando.org", shell=True, check=True)


@pytest.fixture()
def no_peering():
    subprocess.run("kubectl delete customresourcedefinition kopfpeerings.zalando.org", shell=True, check=True)


@pytest.fixture(autouse=True)
def _skip_if_not_explicitly_enabled():
    # Minikube tests are heavy and require a cluster. Skip them by default,
    # so that the contributors can run pytest without initial tweaks.
    if not os.environ.get('E2E'):
        pytest.skip('e2e tests are not explicitly enabled; set E2E env var to enable.')


def pytest_collection_modifyitems(config, items):
    # Put the e2e tests to the end always, since they are a bit lengthy.
    etc = [item for item in items if '/e2e/' not in item.nodeid]
    e2e = [item for item in items if '/e2e/' in item.nodeid]
    items[:] = etc + e2e



import kubernetes


def patch_obj(*, resource, patch, namespace=None, name=None, body=None):
    """
    Patch a resource of specific kind.

    Either the namespace+name should be specified, or the body,
    which is used only to get namespace+name identifiers.

    Unlike the object listing, the namespaced call is always
    used for the namespaced resources, even if the operator serves
    the whole cluster (i.e. is not namespace-restricted).
    """

    if body is not None and (name is not None or namespace is not None):
        raise TypeError("Either body, or name+namespace can be specified. Got both.")

    namespace = body.get('metadata', {}).get('namespace') if body is not None else namespace
    name = body.get('metadata', {}).get('name') if body is not None else name

    api = kubernetes.client.CustomObjectsApi()
    if namespace is None:
        api.patch_cluster_custom_object(
            group=resource.group,
            version=resource.version,
            plural=resource.plural,
            name=name,
            body=patch,
        )
    else:
        api.patch_namespaced_custom_object(
            group=resource.group,
            version=resource.version,
            plural=resource.plural,
            namespace=namespace,
            name=name,
            body=patch,
        )



"""
Kubernetes operator example: all the features at once (for debugging & testing).
"""

import pprint
import time

import kopf

# Marks for the e2e tests (see tests/e2e/test_examples.py):
E2E_CREATE_TIME = 5
E2E_DELETE_TIME = 1
E2E_TRACEBACKS = True


@kopf.on.create('zalando.org', 'v1', 'kopfexamples')
def create_1(body, meta, spec, status, **kwargs):
    children = _create_children(owner=body)

    kopf.info(body, reason='AnyReason')
    kopf.event(body, type='Warning', reason='SomeReason', message="Cannot do something")
    kopf.event(children, type='Normal', reason='SomeReason', message="Created as part of the job1step")

    return {'job1-status': 100}


@kopf.on.create('zalando.org', 'v1', 'kopfexamples')
def create_2(body, meta, spec, status, retry=None, **kwargs):
    wait_for_something()  # specific for job2, e.g. an external API poller

    if not retry:
        # will be retried by the framework, even if it has been restarted
        raise Exception("Whoops!")

    return {'job2-status': 100}


@kopf.on.update('zalando.org', 'v1', 'kopfexamples')
def update(body, meta, spec, status, old, new, diff, **kwargs):
    print('Handling the diff')
    pprint.pprint(list(diff))


@kopf.on.field('zalando.org', 'v1', 'kopfexamples', field='spec.lst')
def update_lst(body, meta, spec, status, old, new, **kwargs):
    print(f'Handling the FIELD = {old} -> {new}')


@kopf.on.delete('zalando.org', 'v1', 'kopfexamples')
def delete(body, meta, spec, status, **kwargs):
    pass


def _create_children(owner):
    return []


def wait_for_something():
    # Note: intentionally blocking from the asyncio point of view.
    time.sleep(1)



import pytest
from asynctest import call

from kopf.k8s.patching import patch_obj


def test_by_name_clustered(client_mock, resource):
    patch = object()
    apicls_mock = client_mock.CustomObjectsApi
    sidefn_mock = apicls_mock.return_value.patch_namespaced_custom_object
    mainfn_mock = apicls_mock.return_value.patch_cluster_custom_object

    res = patch_obj(resource=resource, namespace=None, name='name1', patch=patch)
    assert res is None  # never return any k8s-client specific things

    assert not sidefn_mock.called
    assert mainfn_mock.call_count == 1
    assert mainfn_mock.call_args_list == [call(
        group=resource.group,
        version=resource.version,
        plural=resource.plural,
        name='name1',
        body=patch,
    )]


def test_by_name_namespaced(client_mock, resource):
    patch = object()
    apicls_mock = client_mock.CustomObjectsApi
    sidefn_mock = apicls_mock.return_value.patch_cluster_custom_object
    mainfn_mock = apicls_mock.return_value.patch_namespaced_custom_object

    res = patch_obj(resource=resource, namespace='ns1', name='name1', patch=patch)
    assert res is None  # never return any k8s-client specific things

    assert not sidefn_mock.called
    assert mainfn_mock.call_count == 1
    assert mainfn_mock.call_args_list == [call(
        group=resource.group,
        version=resource.version,
        plural=resource.plural,
        namespace='ns1',
        name='name1',
        body=patch,
    )]


def test_by_body_clustered(client_mock, resource):
    patch = object()
    apicls_mock = client_mock.CustomObjectsApi
    sidefn_mock = apicls_mock.return_value.patch_namespaced_custom_object
    mainfn_mock = apicls_mock.return_value.patch_cluster_custom_object

    body = {'metadata': {'name': 'name1'}}
    res = patch_obj(resource=resource, body=body, patch=patch)
    assert res is None  # never return any k8s-client specific things

    assert not sidefn_mock.called
    assert mainfn_mock.call_count == 1
    assert mainfn_mock.call_args_list == [call(
        group=resource.group,
        version=resource.version,
        plural=resource.plural,
        name='name1',
        body=patch,
    )]


def test_by_body_namespaced(client_mock, resource):
    patch = object()
    apicls_mock = client_mock.CustomObjectsApi
    sidefn_mock = apicls_mock.return_value.patch_cluster_custom_object
    mainfn_mock = apicls_mock.return_value.patch_namespaced_custom_object

    body = {'metadata': {'namespace': 'ns1', 'name': 'name1'}}
    res = patch_obj(resource=resource, body=body, patch=patch)
    assert res is None  # never return any k8s-client specific things

    assert not sidefn_mock.called
    assert mainfn_mock.call_count == 1
    assert mainfn_mock.call_args_list == [call(
        group=resource.group,
        version=resource.version,
        plural=resource.plural,
        namespace='ns1',
        name='name1',
        body=patch,
    )]


def test_raises_when_body_conflicts_with_namespace(client_mock, resource):
    patch = object()
    apicls_mock = client_mock.CustomObjectsApi
    sidefn_mock = apicls_mock.return_value.patch_namespaced_custom_object
    mainfn_mock = apicls_mock.return_value.patch_cluster_custom_object

    body = {'metadata': {'namespace': 'ns1', 'name': 'name1'}}
    with pytest.raises(TypeError):
        patch_obj(resource=resource, body=body, namespace='ns1', patch=patch)

    assert not sidefn_mock.called
    assert not mainfn_mock.called


def test_raises_when_body_conflicts_with_name(client_mock, resource):
    patch = object()
    apicls_mock = client_mock.CustomObjectsApi
    sidefn_mock = apicls_mock.return_value.patch_namespaced_custom_object
    mainfn_mock = apicls_mock.return_value.patch_cluster_custom_object

    body = {'metadata': {'namespace': 'ns1', 'name': 'name1'}}
    with pytest.raises(TypeError):
        patch_obj(resource=resource, body=body, name='name1', patch=patch)

    assert not sidefn_mock.called
    assert not mainfn_mock.called


def test_raises_when_body_conflicts_with_ids(client_mock, resource):
    patch = object()
    apicls_mock = client_mock.CustomObjectsApi
    sidefn_mock = apicls_mock.return_value.patch_namespaced_custom_object
    mainfn_mock = apicls_mock.return_value.patch_cluster_custom_object

    body = {'metadata': {'namespace': 'ns1', 'name': 'name1'}}
    with pytest.raises(TypeError):
        patch_obj(resource=resource, body=body, namespace='ns1', name='name1', patch=patch)

    assert not sidefn_mock.called
    assert not mainfn_mock.called



import collections.abc

import pytest

from kopf.k8s.watching import streaming_watch


class PreventedActualCallError(Exception):
    pass


async def test_watching_over_cluster(resource, client_mock):
    apicls_mock = client_mock.CustomObjectsApi
    cl_list_fn = apicls_mock.return_value.list_cluster_custom_object
    ns_list_fn = apicls_mock.return_value.list_namespaced_custom_object
    cl_list_fn.side_effect = PreventedActualCallError()
    ns_list_fn.side_effect = PreventedActualCallError()

    itr = streaming_watch(
        resource=resource,
        namespace=None,
    )

    assert isinstance(itr, collections.abc.AsyncIterator)
    assert isinstance(itr, collections.abc.AsyncGenerator)

    with pytest.raises(PreventedActualCallError):
        async for _ in itr: pass  # fully deplete it

    assert apicls_mock.called
    assert apicls_mock.call_count == 1

    # Cluster-scoped listing function is used irrelevant of the resource.
    assert not ns_list_fn.called
    assert cl_list_fn.called
    assert cl_list_fn.call_count == 1
    assert cl_list_fn.call_args[1]['group'] == resource.group
    assert cl_list_fn.call_args[1]['version'] == resource.version
    assert cl_list_fn.call_args[1]['plural'] == resource.plural
    assert 'namespace' not in cl_list_fn.call_args[1]


async def test_watching_over_namespace(resource, client_mock):
    apicls_mock = client_mock.CustomObjectsApi
    cl_list_fn = apicls_mock.return_value.list_cluster_custom_object
    ns_list_fn = apicls_mock.return_value.list_namespaced_custom_object
    cl_list_fn.side_effect = PreventedActualCallError()
    ns_list_fn.side_effect = PreventedActualCallError()

    itr = streaming_watch(
        resource=resource,
        namespace='something',
    )

    assert isinstance(itr, collections.abc.AsyncIterator)
    assert isinstance(itr, collections.abc.AsyncGenerator)

    with pytest.raises(PreventedActualCallError):
        async for _ in itr: pass  # fully deplete it

    assert apicls_mock.called
    assert apicls_mock.call_count == 1

    # The scope-relevant listing function is used, depending on the resource.
    assert not cl_list_fn.called
    assert ns_list_fn.called
    assert ns_list_fn.call_count == 1
    assert ns_list_fn.call_args[1]['group'] == resource.group
    assert ns_list_fn.call_args[1]['version'] == resource.version
    assert ns_list_fn.call_args[1]['plural'] == resource.plural
    assert ns_list_fn.call_args[1]['namespace'] == 'something'



import asyncio
import logging

import pytest

import kopf
from kopf.reactor.causation import HANDLER_CAUSES, CREATE, UPDATE, DELETE, RESUME
from kopf.reactor.handling import HandlerFatalError, HandlerRetryError
from kopf.reactor.handling import custom_object_handler


# The extrahandlers are needed to prevent the cycle ending and status purging.
@pytest.mark.parametrize('cause_type', HANDLER_CAUSES)
async def test_fatal_error_stops_handler(
        registry, handlers, extrahandlers, resource, cause_mock, cause_type,
        caplog, assert_logs, k8s_mocked):
    caplog.set_level(logging.DEBUG)
    name1 = f'{cause_type}_fn'

    cause_mock.event = cause_type
    handlers.create_mock.side_effect = HandlerFatalError("oops")
    handlers.update_mock.side_effect = HandlerFatalError("oops")
    handlers.delete_mock.side_effect = HandlerFatalError("oops")
    handlers.resume_mock.side_effect = HandlerFatalError("oops")

    await custom_object_handler(
        lifecycle=kopf.lifecycles.one_by_one,
        registry=registry,
        resource=resource,
        event={'type': 'irrelevant', 'object': cause_mock.body},
        freeze=asyncio.Event(),
    )

    assert handlers.create_mock.call_count == (1 if cause_type == CREATE else 0)
    assert handlers.update_mock.call_count == (1 if cause_type == UPDATE else 0)
    assert handlers.delete_mock.call_count == (1 if cause_type == DELETE else 0)
    assert handlers.resume_mock.call_count == (1 if cause_type == RESUME else 0)

    assert not k8s_mocked.asyncio_sleep.called
    assert k8s_mocked.patch_obj.called

    patch = k8s_mocked.patch_obj.call_args_list[0][1]['patch']
    assert patch['status']['kopf']['progress'] is not None
    assert patch['status']['kopf']['progress'][name1]['failure'] is True
    assert patch['status']['kopf']['progress'][name1]['message'] == 'oops'

    assert_logs([
        "Handler .+ failed with a fatal exception. Will stop.",
    ])


# The extrahandlers are needed to prevent the cycle ending and status purging.
@pytest.mark.parametrize('cause_type', HANDLER_CAUSES)
async def test_retry_error_delays_handler(
        registry, handlers, extrahandlers, resource, cause_mock, cause_type,
        caplog, assert_logs, k8s_mocked):
    caplog.set_level(logging.DEBUG)
    name1 = f'{cause_type}_fn'

    cause_mock.event = cause_type
    handlers.create_mock.side_effect = HandlerRetryError("oops")
    handlers.update_mock.side_effect = HandlerRetryError("oops")
    handlers.delete_mock.side_effect = HandlerRetryError("oops")
    handlers.resume_mock.side_effect = HandlerRetryError("oops")

    await custom_object_handler(
        lifecycle=kopf.lifecycles.one_by_one,
        registry=registry,
        resource=resource,
        event={'type': 'irrelevant', 'object': cause_mock.body},
        freeze=asyncio.Event(),
    )

    assert handlers.create_mock.call_count == (1 if cause_type == CREATE else 0)
    assert handlers.update_mock.call_count == (1 if cause_type == UPDATE else 0)
    assert handlers.delete_mock.call_count == (1 if cause_type == DELETE else 0)
    assert handlers.resume_mock.call_count == (1 if cause_type == RESUME else 0)

    assert not k8s_mocked.asyncio_sleep.called
    assert k8s_mocked.patch_obj.called

    patch = k8s_mocked.patch_obj.call_args_list[0][1]['patch']
    assert patch['status']['kopf']['progress'] is not None
    assert 'failure' not in patch['status']['kopf']['progress'][name1]
    assert 'success' not in patch['status']['kopf']['progress'][name1]
    assert 'delayed' in patch['status']['kopf']['progress'][name1]

    assert_logs([
        "Handler .+ failed with a retry exception. Will retry.",
    ])


# The extrahandlers are needed to prevent the cycle ending and status purging.
@pytest.mark.parametrize('cause_type', HANDLER_CAUSES)
async def test_arbitrary_error_delays_handler(
        registry, handlers, extrahandlers, resource, cause_mock, cause_type,
        caplog, assert_logs, k8s_mocked):
    caplog.set_level(logging.DEBUG)
    name1 = f'{cause_type}_fn'

    cause_mock.event = cause_type
    handlers.create_mock.side_effect = Exception("oops")
    handlers.update_mock.side_effect = Exception("oops")
    handlers.delete_mock.side_effect = Exception("oops")
    handlers.resume_mock.side_effect = Exception("oops")

    await custom_object_handler(
        lifecycle=kopf.lifecycles.one_by_one,
        registry=registry,
        resource=resource,
        event={'type': 'irrelevant', 'object': cause_mock.body},
        freeze=asyncio.Event(),
    )

    assert handlers.create_mock.call_count == (1 if cause_type == CREATE else 0)
    assert handlers.update_mock.call_count == (1 if cause_type == UPDATE else 0)
    assert handlers.delete_mock.call_count == (1 if cause_type == DELETE else 0)
    assert handlers.resume_mock.call_count == (1 if cause_type == RESUME else 0)

    assert not k8s_mocked.asyncio_sleep.called
    assert k8s_mocked.patch_obj.called

    patch = k8s_mocked.patch_obj.call_args_list[0][1]['patch']
    assert patch['status']['kopf']['progress'] is not None
    assert 'failure' not in patch['status']['kopf']['progress'][name1]
    assert 'success' not in patch['status']['kopf']['progress'][name1]
    assert 'delayed' in patch['status']['kopf']['progress'][name1]

    assert_logs([
        "Handler .+ failed with an exception. Will retry.",
    ])



import kubernetes
import pytest

from kopf.config import login, LoginError


@pytest.fixture(autouse=True)
def _auto_clean_kubernetes_client(clean_kubernetes_client):
    pass


def test_direct_auth_works_incluster(mocker):
    # We do not test the client, we assume it works when used properly.
    core_api = mocker.patch.object(kubernetes.client, 'CoreApi')
    load_kube_config = mocker.patch.object(kubernetes.config, 'load_kube_config')
    load_incluster_config = mocker.patch.object(kubernetes.config, 'load_incluster_config')

    login()

    assert load_incluster_config.called
    assert not load_kube_config.called
    assert core_api.called  # to verify that auth worked


def test_direct_auth_works_kubeconfig(mocker):
    # We do not test the client, we assume it works when used properly.
    core_api = mocker.patch.object(kubernetes.client, 'CoreApi')
    load_kube_config = mocker.patch.object(kubernetes.config, 'load_kube_config')
    load_incluster_config = mocker.patch.object(kubernetes.config, 'load_incluster_config')
    load_incluster_config.side_effect = kubernetes.config.ConfigException

    login()

    assert load_incluster_config.called
    assert load_kube_config.called
    assert core_api.called  # to verify that auth worked


def test_direct_auth_fails(mocker):
    # We do not test the client, we assume it works when used properly.
    core_api = mocker.patch.object(kubernetes.client, 'CoreApi')
    load_kube_config = mocker.patch.object(kubernetes.config, 'load_kube_config')
    load_incluster_config = mocker.patch.object(kubernetes.config, 'load_incluster_config')
    load_incluster_config.side_effect = kubernetes.config.ConfigException
    load_kube_config.side_effect = kubernetes.config.ConfigException

    with pytest.raises(LoginError):
        login()

    assert load_incluster_config.called
    assert load_kube_config.called
    assert not core_api.called  # to verify that auth worked


def test_direct_api_fails(mocker):
    # We do not test the client, we assume it works when used properly.
    core_api = mocker.patch.object(kubernetes.client, 'CoreApi')
    load_kube_config = mocker.patch.object(kubernetes.config, 'load_kube_config')
    load_incluster_config = mocker.patch.object(kubernetes.config, 'load_incluster_config')
    core_api.side_effect = kubernetes.client.rest.ApiException(status=401)

    with pytest.raises(LoginError):
        login()

    assert load_incluster_config.called
    assert not load_kube_config.called
    assert core_api.called  # to verify that auth worked


def test_clirun_auth_works_incluster(invoke, mocker, preload, real_run):
    # We do not test the client, we assume it works when used properly.
    core_api = mocker.patch.object(kubernetes.client, 'CoreApi')
    load_kube_config = mocker.patch.object(kubernetes.config, 'load_kube_config')
    load_incluster_config = mocker.patch.object(kubernetes.config, 'load_incluster_config')

    result = invoke(['run'])
    assert result.exit_code == 0

    assert load_incluster_config.called
    assert not load_kube_config.called
    assert core_api.called  # to verify that auth worked


def test_clirun_auth_works_kubeconfig(invoke, mocker, preload, real_run):
    # We do not test the client, we assume it works when used properly.
    core_api = mocker.patch.object(kubernetes.client, 'CoreApi')
    load_kube_config = mocker.patch.object(kubernetes.config, 'load_kube_config')
    load_incluster_config = mocker.patch.object(kubernetes.config, 'load_incluster_config')
    load_incluster_config.side_effect = kubernetes.config.ConfigException

    result = invoke(['run'])
    assert result.exit_code == 0

    assert load_incluster_config.called
    assert load_kube_config.called
    assert core_api.called  # to verify that auth worked


def test_clirun_auth_fails(invoke, mocker, preload, real_run):
    # We do not test the client, we assume it works when used properly.
    core_api = mocker.patch.object(kubernetes.client, 'CoreApi')
    load_kube_config = mocker.patch.object(kubernetes.config, 'load_kube_config')
    load_incluster_config = mocker.patch.object(kubernetes.config, 'load_incluster_config')
    load_incluster_config.side_effect = kubernetes.config.ConfigException
    load_kube_config.side_effect = kubernetes.config.ConfigException

    result = invoke(['run'])
    assert result.exit_code != 0
    assert 'neither in-cluster, nor via kubeconfig' in result.stdout

    assert load_incluster_config.called
    assert load_kube_config.called
    assert not core_api.called  # to verify that auth worked


def test_clirun_api_fails(invoke, mocker, preload, real_run):
    # We do not test the client, we assume it works when used properly.
    core_api = mocker.patch.object(kubernetes.client, 'CoreApi')
    load_kube_config = mocker.patch.object(kubernetes.config, 'load_kube_config')
    load_incluster_config = mocker.patch.object(kubernetes.config, 'load_incluster_config')
    core_api.side_effect = kubernetes.client.rest.ApiException(status=401)

    result = invoke(['run'])
    assert result.exit_code != 0
    assert 'Please login or configure the tokens' in result.stdout

    assert load_incluster_config.called
    assert not load_kube_config.called
    assert core_api.called  # to verify that auth worked




import asyncio
import logging

import click
import kubernetes
import kubernetes.client.rest
import urllib3.exceptions

logger = logging.getLogger(__name__)
format = '[%(asctime)s] %(name)-20.20s [%(levelname)-8.8s] %(message)s'


class LoginError(click.ClickException):
    """ Raised when the operator cannot login to the API. """


def login():
    """
    Login the the Kubernetes cluster, locally or remotely.
    """

    # Configure the default client credentials for all possible environments.
    try:
        kubernetes.config.load_incluster_config()  # cluster env vars
        logger.debug("configured in cluster with service account")
    except kubernetes.config.ConfigException as e1:
        try:
            kubernetes.config.load_kube_config()  # developer's config files
            logger.debug("configured via kubeconfig file")
        except kubernetes.config.ConfigException as e2:
            raise LoginError(f"Cannot authenticate neither in-cluster, nor via kubeconfig.")

    # Make a sample API call to ensure the login is successful,
    # and convert some of the known exceptions to the CLI hints.
    try:
        api = kubernetes.client.CoreApi()
        api.get_api_versions()
    except urllib3.exceptions.HTTPError as e:
        raise LoginError("Cannot connect to the Kubernetes API. "
                         "Please configure the cluster access.")
    except kubernetes.client.rest.ApiException as e:
        if e.status == 401:
            raise LoginError("Cannot authenticate to the Kubernetes API. "
                             "Please login or configure the tokens.")
        else:
            raise


def configure(debug=None, verbose=None, quiet=None):
    log_level = 'DEBUG' if debug or verbose else 'WARNING' if quiet else 'INFO'

    logger = logging.getLogger()
    handler = logging.StreamHandler()
    formatter = logging.Formatter(format)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(log_level)

    # Configure the Kubernetes client defaults according to our settings.
    config = kubernetes.client.configuration.Configuration()
    config.logger_format = format
    config.logger_file = None  # once again after the constructor to re-apply the formatter
    config.debug = debug
    kubernetes.client.configuration.Configuration.set_default(config)

    # Kubernetes client is as buggy as hell: it adds its own stream handlers even in non-debug mode,
    # does not respect the formatting, and dumps too much of the low-level info.
    if not debug:
        logger = logging.getLogger("urllib3")
        del logger.handlers[1:]  # everything except the default NullHandler

    # Prevent the low-level logging unless in the debug verbosity mode. Keep only the operator's messages.
    logging.getLogger('urllib3').propagate = debug
    logging.getLogger('asyncio').propagate = debug
    logging.getLogger('kubernetes').propagate = debug

    loop = asyncio.get_event_loop()
    loop.set_debug(debug)



"""
All the functions to keep track of the last seen state.

The "state" is a snapshot of meaningful fields, which must be tracked
to identify the actual changes on the object (or absence of such).

Used in the handling routines to check if there were significant changes at all
(i.e. not the internal and system changes, like the uids, links, etc),
and to get the exact per-field diffs for the specific handler functions.

Conceptually similar to how ``kubectl apply`` stores the applied state
on any object, and then uses that for the patch calculation:
https://kubernetes.io/docs/concepts/overview/object-management-kubectl/declarative-config/
"""

import copy
import json

from kopf.structs import diffs

LAST_SEEN_ANNOTATION = 'kopf.zalando.org/last-handled-configuration'
""" The annotation name for the last stored state of the resource. """


def get_state(body):
    """
    Extract only the relevant fields for the state comparisons.
    """

    # Always use a copy, so that future changes do not affect the extracted state.
    body = copy.deepcopy(body)

    # Remove the system fields, keeping the potentially useful, user-oriented fields/data.
    if LAST_SEEN_ANNOTATION in body.get('metadata', {}).get('annotations', {}):
        del body['metadata']['annotations'][LAST_SEEN_ANNOTATION]
    if 'kubectl.kubernetes.io/last-applied-configuration' in body.get('metadata', {}).get('annotations', {}):
        del body['metadata']['annotations']['kubectl.kubernetes.io/last-applied-configuration']
    if 'finalizers' in body.get('metadata', {}):
        del body['metadata']['finalizers']
    if 'deletionTimestamp' in body.get('metadata', {}):
        del body['metadata']['deletionTimestamp']
    if 'creationTimestamp' in body.get('metadata', {}):
        del body['metadata']['creationTimestamp']
    if 'selfLink' in body.get('metadata', {}):
        del body['metadata']['selfLink']
    if 'uid' in body.get('metadata', {}):
        del body['metadata']['uid']
    if 'resourceVersion' in body.get('metadata', {}):
        del body['metadata']['resourceVersion']
    if 'generation' in body.get('metadata', {}):
        del body['metadata']['generation']
    if 'kopf' in body.get('status', {}):
        del body['status']['kopf']

    # Cleanup the parent structs if they have become empty, for consistent state comparison.
    if 'annotations' in body.get('metadata', {}) and not body['metadata']['annotations']:
        del body['metadata']['annotations']
    if 'metadata' in body and not body['metadata']:
        del body['metadata']
    if 'status' in body and not body['status']:
        del body['status']
    return body


def has_state(body):
    annotations = body.get('metadata', {}).get('annotations', {})
    return LAST_SEEN_ANNOTATION in annotations


def is_state_changed(body):
    # TODO: make it more efficient, so that the dicts are not rebuilt locally every time.
    old = retreive_state(body)
    new = get_state(body)
    return old != new


def get_state_diffs(body):
    old = retreive_state(body)
    new = get_state(body)
    return old, new, diffs.diff(old, new)


def retreive_state(body):
    state_str = body.get('metadata', {}).get('annotations', {}).get(LAST_SEEN_ANNOTATION, None)
    state_obj = json.loads(state_str) if state_str is not None else None
    return state_obj


def refresh_state(*, body, patch):
    state = get_state(body)
    patch.setdefault('metadata', {}).setdefault('annotations', {})[LAST_SEEN_ANNOTATION] = json.dumps(state)
